#include "dserv.h"


//////// Directory settings ////////

static directorySettings ds;


//////// Direct updates ////////

bool nameMatch(const char * a, const char * b)
{
	Uint32 lena = limit(strlen(a), 21);		// Only 21 characters are visible on the server list
	Uint32 lenb = limit(strlen(b), 21);
	Uint32 len = (lena > lenb) ? lena : lenb;

	Uint32 capCount = 0;

	for (Uint32 i = 0; i < len; i++)
	{
		char aa = a[i],
			 bb = b[i];

		if ((aa >= 'A') && (aa <= 'Z'))
		{
			aa -= 'A' - 'a';	// Case-insensitive match
		}

		if (aa == 'l')
			aa = 'I';

		if ((bb >= 'A') && (bb <= 'Z'))
		{
			bb -= 'A' - 'a';
			capCount++;			// (char*)b is the incoming string
		}

		if (bb == 'l')
			bb = 'I';			// Catch simple substitution hacks

		if (aa != bb)
			return false;

		if (bb == ' ')			// Trim down to the network prefix if applicable
		{
			switch (i)
			{
			case 2:
				if (capCount)
					return true;
				break;
			case 3:
			case 4:
				if (capCount >= 2)
					return true;
			};
		}
	}

	return true;
}

bool passMatch(const char * a, const char * b)
{
	Uint32 len = limit(strlen(a), 15);		// Only 15 characters are available for zone passwords
	Uint32 lenb = limit(strlen(b), 15);
	if (len != lenb) return false;

	for (Uint32 i = 0; i < len; i++)
	{
		if (a[i] != b[i])
			return false;
	}

	return true;
}

bool validName(const char * name)
{
	if (strlen(name) == 0)
		return true;			// Allow zones with no names

	bool seenSpace = true;		// No leading spaces

	for (Uint32 i = 0; i < strlen(name); i++)
	{
		char c = name[i];

		if (c == ' ')
		{
			if (seenSpace)		// No more than one space in a row
				return false;
			seenSpace = true;
		}
		else
		{
			seenSpace = false;

			if ((c < ' ') || (c > '~'))
				return false;	// No non-printable characters
		}
	}

	if (seenSpace)
		return false;			// No trailing spaces

	return true;
}

bool validUpdate(char * msg, Uint32 len)
{
	/*
		Because this system is EXTREMELY simplistic and doesn't
		allow for host authentication via two-way protocol, we
		must be as picky as possible about incoming updates.
	*/

	if (len >= 0x5F)				// Fixed-length fields are present
	if (getLong(msg, 0) == 0)		// The IP field must be NULL
	if (msg[len - 1] == 0)			// The comments field has a trailing zero
	if (msg[0x2D] == 0)				// The name field has a trailing zero
	if (msg[0x3D] == 0)				// The password field has a trailing zero
	if ((getShort(msg, 8) == 0) ||
		(getShort(msg, 8) == 1))	// Scoring field must be 0 or 1
	{	// Check field contents
		int i;

		for (i = 62; i < 62 + 32; i++)
		{	// Make sure the NULL field is NULL
			if (msg[i]) return false;
		}

		// Make sure the name field is printable
		if (validName(&msg[14]))
			return true;
	}

	return false;
}


//////// Password system ////////

bool passwordList::validate(char * name, char * pass)
{
	_listnode <passwordEntry> * parse = plist.head;

	while (parse)
	{
		passwordEntry * e = parse->item;

		if (nameMatch(e->name, name))
		{
			if (e->validate(pass))
				return true;

			return false;
		}

		parse = parse->next;
	}

	passwordEntry * e = new passwordEntry(name, pass);
	if (e == NULL) return false;

	plist.append(e);
	return true;
}


//////// Password entry ////////

passwordEntry::passwordEntry(char * nname, char * npass)
{
	for (Uint32 i = 0; i < 32; i++)
		name[i] = 0;

	for (i = 0; i < 16; i++)
		pass[i] = 0;

	MEMCPY(nname, name, limit(strlen(nname), 31));
	MEMCPY(npass, pass, limit(strlen(npass), 15));

	lastTime = GetTickCount();
}

bool passwordEntry::validate(char * tpass)
{
	if (passMatch(pass, tpass))
	{
		lastTime = GetTickCount();

		return true;
	}

	if (timeout())
	{
		for (int i = 0; i < 16; i++)
			pass[i] = 0;

		MEMCPY(tpass, pass, limit(strlen(tpass), 15));

		lastTime = GetTickCount();

		return true;
	}

	return false;
}

bool passwordEntry::timeout()
{
	return GetTickCount() - lastTime > ds.ReclaimDelay;
}


//////// Server system ////////

void zoneList::addZone(serverEntry * se)
{
	zlist.append(se);
}

zoneList::zoneList()
{
	target = NULL;
}

serverEntry * zoneList::find(Uint32 ip, Uint16 port)
{
	_listnode <serverEntry> * parse = zlist.head;

	while (parse)
	{
		serverEntry * se = parse->item;

		if (se->validate(ip, port))
			return se;

		parse = parse->next;
	}

	return NULL;
}

serverEntry * zoneList::getNext()
{
	if (target == NULL)
	{
		target = zlist.head;
		if (target == NULL)
			return NULL;
	}

	target = target->next;

	if (target == NULL)
		target = zlist.head;

	if (target)
		return target->item;
	else
		return NULL;
}

void zoneList::killZone(serverEntry * se)
{
	if (target == NULL)
		return;

	if (target->item == se)
		target = target->last;

	zlist.kill(se, false);
}

void zoneList::updateZone(Uint32 ip, Uint16 port, Uint16 population, Uint16 scoring, Uint32 version, char * name, char * comment)
{
	serverEntry * se = find(ip, port);

	if (scoring != 0)
		scoring = 1;

	if (se)
	{
		se->scoring = scoring;
		se->version = version;
		se->set(name, comment);
	}
	else
	{
		if (validName(name) == false)
			return;

		se = new serverEntry(ip, port, population, scoring, version, name, comment);
		if (se)
		{
			zlist.append(se);
			printf("Adding : %s\n", se->name);
		}
	}
}


//////// Server entry ////////

serverEntry::serverEntry()
{
	pinging = false;
	highPing = ds.InitialHighPing;
	dropped = 0;
	lastTime = 0;
}

serverEntry::serverEntry(Uint32 nip, Uint16 nport, Uint16 npopulation, Uint16 nscoring, Uint32 nversion, char * nname, char * ncomment)
{
	ip = nip;
	port = nport;
	population = npopulation;
	scoring = nscoring;
	version = nversion;

	set(nname, ncomment);

	pinging = false;
	highPing = ds.InitialHighPing;
	dropped = 0;
	lastTime = 0;
}

bool serverEntry::validate(Uint32 tip, Uint16 tport)
{
	return ((ip == tip) && (port == tport));
}

void serverEntry::set(char * nname, char * ncomment)
{
	for (int i = 0; i < 32; i++)
		name[i] = 0;
	MEMCPY(nname, name, limit(strlen(nname), 31));
	for (i = 0; i < 256; i++)
		comment[i] = 0;
	MEMCPY(ncomment, comment, limit(strlen(ncomment), 255));
}

void serverEntry::startPinging()
{
	pinging = true;
	dropped = 0;
	lastTime = GetTickCount();
}

void serverEntry::gotPong(Uint32 pingTime, Uint32 npopulation)
{
	Uint32 time = GetTickCount() - pingTime;
	population = (Uint16)npopulation;

	if ((time > highPing) && (time <= 1000))
	{
		highPing = time;
	}

	pinging = false;
}

bool serverEntry::resendPing()
{
	if (GetTickCount() - lastTime > highPing)
	{
		lastTime = GetTickCount();
		dropped++;
		return true;
	}

	return false;
}

bool serverEntry::timeout()
{
	return (dropped > ds.DropLimit);
}


//////// Directory server ////////

directoryServer::directoryServer() : upload((Uint16)ds.MaxHosts, 4990)
{
	update.create(4991);
	sweep.create(0);

	state = iteration = lastDownload = 0;
	lastWrite = lastTime = GetTickCount();

	ds.reloadSettings();

	busy = false;
}

directoryServer::~directoryServer()
{
	upload.massDisconnect();
	download.massDisconnect();
}

void directoryServer::sendPing(Uint32 ip, Uint16 port)
{
	char msg[4];
	*(Uint32*)msg = GetTickCount();
	sweep.set(ip, port);
	sweep.send(msg, 4);
}

void directoryServer::sendPong(Uint32 ip, Uint16 port, Uint32 pingTime)
{
	char msg[8];
	*(Uint32*)msg = zlist.zlist.total;
	*(Uint32*)&msg[4] = pingTime;
	sweep.set(ip, port);
	sweep.send(msg, 8);
}

clusterMessage * directoryServer::makeZoneList(Uint32 min)
{
	_listnode <serverEntry> * parse = zlist.zlist.head;

	chunkBuffer cb;
	cb.setLimit(50000);

	char pckt[1];
	pckt[0] = 0x01;
	cb.addMessage(pckt, 1);

	char nullBuffer[32];
	memset(nullBuffer, 0, 32);

	while (parse)
	{
		serverEntry * se = parse->item;

		if (se->population >= min)
		{
			cb.addMessage((char*)&se->ip, 4);
			cb.addMessage((char*)&se->port, 2);
			cb.addMessage((char*)&se->population, 2);
			cb.addMessage((char*)&se->scoring, 2);
			cb.addMessage((char*)&se->version, 4);
			cb.addMessage(se->name, 32);
			cb.addMessage(nullBuffer, 32);
			cb.addMessage(se->comment, strlen(se->comment) + 1);
		}

		parse = parse->next;
	}

	clusterMessage * m = new clusterMessage(cb.buffer, cb.currentLength);

	printf("---Uploading server list\n");

	return m;
}

void directoryServer::resetPings()
{
	_listnode <serverEntry> * parse = zlist.zlist.head;
	while (parse)
	{
		parse->item->pinging = false;
		parse = parse->next;
	}
}

void directoryServer::writeStats()
{
	// ip, port, population, scoring, name, comments
	// 63.198.238.211	2000	0	1	CAT Test Zone	SSBot3 testing grounds

	printf("---Writing stats\n");

	INADDR addr;

	fileAccess fa;
	_unlink(ds.StatsFile);
	fa.open(ds.StatsFile);

	char buffer[520];

	_listnode <serverEntry> * parse = zlist.zlist.head;

	while (parse)
	{
		serverEntry * se = parse->item;

		addr.set(se->ip, 0);

		sprintf(buffer, "%s\t%i\t%i\t%i\t%s\t%s\r\n", addr.getString(),
													  se->port,
													  se->population,
													  se->scoring,
													  se->name,
													  se->comment);

		fa.writeStuff(buffer, strlen(buffer));

		parse = parse->next;
	}

	fa.close();
}

void directoryServer::doEvents()
{
	// Ignore timeouts if there's CPU lag
	if (iteration == 0)
	{
		iteration = GetTickCount();
	}
	else if (GetTickCount() - iteration > ds.SlowIteration)
	{
		printf("Slow iteration warning: %i ms\n", GetTickCount() - iteration);

		resetPings();
	}

	iteration = GetTickCount();

	// Downloads from other directory servers
	download.doEvents();

	// Uploads to SubSpace clients
	upload.doEvents();

	UDPPacket * p;

	// Direct zone updates
	p = update.poll();

	if (p)
	{
		if (validUpdate(p->msg, p->len))
		{
			if (plist.validate(p->msg + 0x0e, p->msg + 0x2e))
			{
				zlist.updateZone(p->src.ip,				// ip
								 getShort(p->msg, 4),	// port
								 getShort(p->msg, 6),	// population
								 getShort(p->msg, 8),	// scoring
								 getLong(p->msg, 10),	// version
								 p->msg + 0x0e,			// name
								 p->msg + 0x5e);		// comment

				printf("---Update : %s\n", p->msg + 0x0e);
			}
			else
			{
				printf("---Bad password from %s: %s\n", p->src.getString(), p->msg + 0x0e);
			}
		}
		else if (p->len == 4)
		{
			sendPong(p->src.ip, p->src.getPort(), getLong(p->msg, 0));
		}
		else
		{
			printf("---Invalid message on direct update port from %s\n", p->src.getString());
		}

		delete p;
	}

	// Ping-sweep responses
	p = sweep.poll();

	if (p)
	{
		if (p->len == 8)
		{
			serverEntry * se = zlist.find(p->src.ip, p->src.getPort() - 1);
			if (se)
			{
				se->gotPong(getLong(p->msg, 4), getLong(p->msg, 0));
			}
		}
		else
		{
			printf("---Invalid message on ping sweep port\n");
		}

		delete p;
	}

	// Check if we need to download the server list again
	if (download.list.total)
	{
		lastDownload = GetTickCount();

		if (busy == false)
		{
			// Clear pings
			resetPings();

			busy = true;
		}
	}
	else
	{
		if (GetTickCount() - lastDownload > ds.DownloadTime)
		{
			printf("---Downloading server list\n");
			download.connectHost(ds.TrustedHosts[state++], 4990);

			if (state == ds.TotalHosts)
				state = 0;
		}
		else if (upload.list.total)
		{
			if (busy == false)
			{
				// Clear pings
				resetPings();

				busy = true;
			}
		}
		else
		{
			busy = false;
		}
	}

	// Check busy signal
	if (busy == false)
	{
		// Resend ping requests to unresponsive hosts
		_listnode <serverEntry> * parse = zlist.zlist.head;

		while (parse)
		{
			serverEntry * se = parse->item;
			parse = parse->next;

			if (se->pinging)
			{
				if (se->timeout())
				{
					printf("Timeout : %s\n", se->name);
					zlist.killZone(se);
				}
				else
				{
					if (se->resendPing())
					{
						sendPing(se->ip, se->port + 1);
					}
				}
			}
		}

		// Start pinging a new host
		if (GetTickCount() - lastTime > ds.SweepTime)
		{
			serverEntry * se = zlist.getNext();

			if (se && (se->pinging == false))
			{
				lastTime = GetTickCount();
				se->startPinging();
				sendPing(se->ip, se->port + 1);
			}
		}

		// Write stats
		if (GetTickCount() - lastWrite > ds.StatsWriteMinutes)
		{
			if (ds.StatsFile[0]) writeStats();

			lastWrite = GetTickCount();
		}
	}
}
